/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Axis2BCListener.java 154 19 avr. 2006 wjoseph $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import static org.ow2.petals.binding.soap.Constants.SOAP.FAULT_CLIENT;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.receivers.AbstractMessageReceiver;
import org.apache.axis2.util.MessageContextBuilder;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;

/**
 * The message receiver used by the binding component. The SOAP message is
 * forwarded to the JBI endpoint.
 * 
 * @version $Id$
 * @since Petals 1.0
 * @author alouis,wjoseph,chamerling - EBM Websourcing
 * 
 */
public class PetalsReceiver extends AbstractMessageReceiver {

    protected Map<String, RequestProcessor> requestProcessors;

    protected AbstractComponent component;

    protected DeliveryChannel channel;

    protected Logger logger;

    /**
     * Creates a new instance of PetalsReceiver
     * 
     * @param context
     * @param channel
     * @param bindingSUM
     * @param log
     */
    public PetalsReceiver(AbstractComponent component, DeliveryChannel channel, Logger logger) {
        super();
        this.logger = logger;
        this.component = component;
        this.channel = channel;
        this.requestProcessors = new HashMap<String, RequestProcessor>();

        this.initRequestProcessors();
    }

    /**
     * Init the request processors
     * 
     */
    protected void initRequestProcessors() {
        this.addProcessor(new RESTProcessor(component, channel, logger));
        this.addProcessor(new SOAPProcessor(component, channel, logger));
    }

    /**
     * Add a processor to the processors list
     * 
     * @param processor
     */
    protected void addProcessor(final RequestProcessor processor) {
        this.requestProcessors.put(processor.getProcessorName(), processor);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.apache.axis2.engine.MessageReceiver#invokeBusinessLogic(org.apache.axis2.context.MessageContext)
     */
    public final void invokeBusinessLogic(final MessageContext msgContext) throws AxisFault {
        logger.log(Level.FINE, "Receiving a message from Axis2 engine");

        // Performance notification creation
        PerformanceNotification performanceNotification = this.component
                .createPerformanceNotificationReceivingFromOutside(msgContext.getTo().getAddress(),
                        null);

        try {
            // FIX NPE (Axis2 bug)
            Parameter param = new Parameter(
                    Constants.Configuration.SEND_STACKTRACE_DETAILS_WITH_FAULTS, Boolean.TRUE);
            if (msgContext.getAxisOperation() != null) {
                msgContext.getAxisOperation().addParameter(param);
            } else if (msgContext.getAxisService() != null) {
                msgContext.getAxisService().addParameter(param);
            } else if (msgContext.getAxisServiceGroup() != null) {
                msgContext.getAxisServiceGroup().addParameter(param);
            }

            MessageContext outMsgContext = MessageContextBuilder
                    .createOutMessageContext(msgContext);
            outMsgContext.getOperationContext().addMessageContext(outMsgContext);

            // send message to JBI endpoint
            this.process(msgContext, outMsgContext, performanceNotification);

            // send the response if required
            if (outMsgContext.getEnvelope() != null) {
                AxisEngine.send(outMsgContext);
            }
        } finally {
            // Performance notification sending
            this.component.sendPerformanceNotification(performanceNotification);
        }
    }

    /**
     * Sends the SOAPEnvelope of the request message directly to a JBI service
     * and operation on the JBI container. The response message from the JBI
     * service invocation is transmitted as an outgoing message.
     * 
     * @param inMessage
     *            MessageContext for incoming message.
     * @param outMessage
     *            MessageContect fot outgoing message.
     * @param performanceNotification
     *            The performance notification whose the UID will be forwarded
     *            to the service provider.
     * @throws AxisFault
     *             error while processing request, send the JBI Exchange,
     *             processing the JBI response, or if the JBI response is a
     *             Fault.
     */
    protected void process(final MessageContext inContext, final MessageContext outContext,
            final PerformanceNotification performanceNotification) throws AxisFault {

        if (!(inContext.getAxisService() instanceof PetalsAxisService)) {
            throw new AxisFault("Can not retrieve a valid PEtALS Axis service from context",
                    FAULT_CLIENT);
        }

        RequestProcessor processor = getProcessor(inContext);
        if (processor == null) {
            throw new AxisFault("Can not process the incoming request, "
                    + "message processor has not been found for this type of request");
        }

        processor
                .process(inContext, outContext, getSOAPFactory(inContext), performanceNotification);
    }

    /**
     * Get the processor for the given message context.
     * 
     * @param context
     * @return the processor of the SOAP one which is the default one if no
     *         valid processor has been found.
     */
    protected RequestProcessor getProcessor(MessageContext context) {
        RequestProcessor result = null;
        if (context.isDoingREST()) {
            result = getRESTProcessor();
        } else {
            result = getSOAPProcessor();
        }
        return result;
    }

    /**
     * Get the REST processor
     * 
     * @return
     */
    protected RequestProcessor getRESTProcessor() {
        return requestProcessors.get("REST");
    }

    /**
     * Get the SOAP processor
     * 
     * @return
     */
    protected RequestProcessor getSOAPProcessor() {
        return requestProcessors.get("SOAP");
    }

}
