/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import static org.ow2.petals.binding.soap.Constants.SOAP.ERROR_WRONG_MESSAGE_STATUS;
import static org.ow2.petals.binding.soap.Constants.SOAP.FAULT_SERVER;
import static org.ow2.petals.component.framework.Constants.Component.MEP;
import static org.ow2.petals.component.framework.Constants.MEPConstants.IN_ONLY_PATTERN;
import static org.ow2.petals.component.framework.Constants.MEPConstants.IN_OPTIONAL_OUT_PATTERN;
import static org.ow2.petals.component.framework.Constants.MEPConstants.IN_OUT_PATTERN;
import static org.ow2.petals.component.framework.Constants.MEPConstants.ROBUST_IN_ONLY_PATTERN;

import java.net.URI;
import java.security.Principal;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.MessageContext;
import org.apache.ws.security.WSSecurityEngineResult;
import org.apache.ws.security.handler.WSHandlerConstants;
import org.apache.ws.security.handler.WSHandlerResult;
import org.ow2.petals.binding.soap.listener.SUPropertiesHelper;
import org.ow2.petals.binding.soap.util.Marshaller;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.Constants.MEPShortcuts;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.util.ExchangeImpl;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.w3c.dom.Document;

/**
 * 
 * Created on 14 févr. 08
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since
 * 
 */
public abstract class RequestProcessor {

    protected Logger logger;

    protected DeliveryChannel channel;

    protected AbstractComponent component;

    protected MessageExchangeFactory exchangeFactory;

    protected Marshaller marshaller = Marshaller.getInstance();

    /**
     * 
     * @param component
     * @param channel
     * @param log
     */
    public RequestProcessor(AbstractComponent component, DeliveryChannel channel, Logger log) {
        this.component = component;
        this.channel = channel;
        this.logger = log;
        this.exchangeFactory = channel.createExchangeFactory();
    }

    /**
     * 
     * @param inContext
     * @param outContext
     * @param performanceNotification
     * @throws AxisFault
     */
    protected abstract void process(final MessageContext inContext,
            final MessageContext outContext, final SOAPFactory soapfactory,
            final PerformanceNotification performanceNotification) throws AxisFault;

    /**
     * 
     * @return
     */
    // protected abstract QName getJBIOperation(final MessageContext inContext);
    /**
     * Process the JBI response. <br>
     * According to the response status :
     * <li>
     * <ul>
     * do nothin for DONE,
     * </ul>
     * <ul>
     * throws an AxisFault for ERROR,
     * </ul>
     * <ul>
     * set outContext body to a FAULT,
     * </ul>
     * <ul>
     * set outContext body to the response.
     * </ul>
     * </li>
     * 
     * @param exchange
     *            JBI response
     * @param outMessage
     *            set to FAULT or OUT JBI response
     * @param factory
     *            soap factory used to create body response
     * @throws AxisFault
     *             with ERROR JBI status, or problem while creating response
     */
    protected void processJBIResponse(MessageExchange exchange, MessageContext outMessage,
            SOAPFactory factory) throws AxisFault {
        if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
            // exchange DONE, nothing to do, no soap response

        } else if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
            // exchange ERROR: throws an axis fault
            throw new AxisFault(FAULT_SERVER, exchange.getError());

        } else if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
            if (exchange.getFault() != null) {
                // exchange FAULT : return a fault message
                logger.warning("Got a fault on JBI response");
                try {
                    SOAPEnvelope envelope = marshaller.createSOAPEnvelopeFromNM(factory, exchange
                            .getFault());
                    outMessage.setEnvelope(envelope);
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, FAULT_SERVER, e);
                }

            } else {
                // Get the output message if the message is an instance of out
                try {
                    NormalizedMessage nm = null;
                    if (exchange instanceof InOptionalOut) {
                        InOptionalOut mex = (InOptionalOut) exchange;
                        nm = mex.getOutMessage();
                    } else if (exchange instanceof InOut) {
                        InOut mex = (InOut) exchange;
                        nm = mex.getOutMessage();
                    }

                    SOAPEnvelope envelope = marshaller.createSOAPEnvelopeFromNM(factory, nm);
                    outMessage.setEnvelope(envelope);

                    marshaller.copyJBIAttachmentsToSOAP(nm, outMessage);
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, FAULT_SERVER, e);
                }
            }
            // close the messageExchange with the JBI NMR
            try {
                exchange.setStatus(ExchangeStatus.DONE);
                channel.send(exchange);
            } catch (MessagingException e) {
                String errorMsg = "Error while closing JBI MessageExchange.";
                logger.log(Level.SEVERE, errorMsg, e);
                throw new AxisFault(errorMsg, FAULT_SERVER, e);
            }

        } else {
            throw new AxisFault(ERROR_WRONG_MESSAGE_STATUS + " " + exchange.getStatus().toString());
        }
    }

    /**
     * Get the pattern of the given operation. First, try to get pattern from
     * the WSDL definition of the service. If the result is null, try to get the
     * pattern from the extensions. Note that in this case, the pattern is not
     * defined for the operation but to all of them.
     * 
     * @param operation
     * @param axisService
     * @return
     */
    protected final MEPConstants getPattern(QName operation, PetalsAxisService axisService) {
        MEPConstants mep = retrievePatternFromDefinition(operation, axisService);
        if (mep == null) {
            mep = retrievePatternFromExtensions(axisService.getConsumesCDKExtensions());
        }
        return mep;
    }

    /**
     * Retrieve the MEP from the axisService
     * 
     * @param operation
     * @param axisService
     * @return
     */
    protected final MEPConstants retrievePatternFromDefinition(QName operation,
            PetalsAxisService axisService) {
        MEPConstants mep = null;
        logger.fine("Get MEP for operation " + operation);

        try {
            // get from cache
            URI mepURI = axisService.getOperationMep().get(operation);
            if (mepURI == null) {
                // retrieve MEP from definition (done on first call)
                Document definition = axisService.getServiceDefinition();
                if (definition != null) {
                    // WsdlHelper.getOperation(definition, bodyName,
                    // serviceName,
                    // interfaceName, endpointName);
                }
            }
        } catch (Exception e) {
        }
        return mep;
    }

    /**
     * Retrieve the MEP from the extensions.
     * 
     * @param extensions
     * @return MEP specified in extensions or IN_OPTIONAL_OUT one if not found
     */
    protected final MEPConstants retrievePatternFromExtensions(final Extensions extensions) {
        MEPConstants mep = null;
        String patternShortcut = extensions.getPetalsExtensions().getProperty(MEP);

        if (!StringHelper.isNullOrEmpty(patternShortcut)) {
            if (MEPShortcuts.ROBUST_IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = ROBUST_IN_ONLY_PATTERN;
            } else if (MEPShortcuts.IN_OPTIONAL_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = IN_OPTIONAL_OUT_PATTERN;
            } else if (MEPShortcuts.IN_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = IN_OUT_PATTERN;
            } else if (MEPShortcuts.IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = IN_ONLY_PATTERN;
            } else {
                mep = IN_OPTIONAL_OUT_PATTERN;
            }
        } else {
            mep = IN_OPTIONAL_OUT_PATTERN;
        }
        return mep;
    }

    /**
     * An utility method creating a message exchange matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return a message exchange matching the given pattern
     * @throws MessagingException
     *             if the message exchange can't be created or if the specified
     *             pattern isn't valid
     */
    protected MessageExchange createMessageExchange(final MEPConstants mep)
            throws MessagingException {
        if (mep == null) {
            throw new IllegalArgumentException("'mep' parameter must be non null");
        }
        MessageExchange exchange = null;
        if (mep.equals(MEPConstants.IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOnlyExchange();
        } else if (mep.equals(MEPConstants.IN_OPTIONAL_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOptionalOutExchange();
        } else if (mep.equals(MEPConstants.IN_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOutExchange();
        } else if (mep.equals(MEPConstants.ROBUST_IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory().createRobustInOnlyExchange();
        } else {
            throw new MessagingException("Given mep doesn't exist");
        }
        return exchange;
    }

    /**
     * An util method creating a message exchange matching a pattern specified
     * in CDKExtensions. Valid patterns are : "in-only", "robust-in-only",
     * "in-opt-out" or "in-out"
     * 
     * @param extensions
     *            a {@link CDKExtensions}, must not be null
     * @return a message exchange matching the pattern given in extensions
     * @throws MessagingException
     *             if the message exchange can't be created, if the specified
     *             pattern isn't valid or if no pattern specified in extensions
     */
    protected MessageExchange createMessageExchange(final Extensions extensions)
            throws MessagingException {
        if (extensions == null) {
            throw new NullPointerException("'extensions' parameter must be non null");
        }
        MEPConstants mep = retrievePatternFromExtensions(extensions);
        return createMessageExchange(mep);
    }

    /**
     * Return the JBI ServiceEndpoint according to : <br>{
     * axisService.getTargetNamespace() } axisService.getName() <br>
     * axisService.getEndpoint()
     * 
     * @param consumes
     * @return the corresponding endpoint, null if no JBI endpoint has been
     *         found
     */
    protected ServiceEndpoint findJBIEndpoint(final Consumes consumes) {
        ServiceEndpoint endpoint = null;
        if (consumes.getEndpointName() != null && consumes.getServiceName() != null) {
            endpoint = component.getContext().getEndpoint(consumes.getServiceName(),
                    consumes.getEndpointName());
        }
        return endpoint;
    }

    /**
     * Create the JBI message exchange
     * 
     * @param inContext
     * @param operationName
     * @param axisService
     * @param performanceNotification
     *            The performance notification whose the UID will be forwarded
     *            to the service provider.
     * @return
     * @throws AxisFault
     */
    protected MessageExchange createMessageExchange(MessageContext inContext,
            final QName operationName, final PetalsAxisService axisService,
            PerformanceNotification performanceNotification) throws AxisFault {
        ExchangeImpl exchangeWrapper = null;

        // retrieve MEP for operation
        // MEPConstants mep = getPattern(operationName, axisService);
        try {

            final Source source = marshaller.createSourceFromSOAPEnvelope(inContext.getEnvelope(),
                    SUPropertiesHelper.retrieveRemoveRootElement(axisService
                            .getConsumesCDKExtensions().getPetalsExtensions(), false));

            // create the message exchange from the extensions. If the MEP is
            // not defined, default will be IN_OPTIONAL_OUT one.
            MessageExchange exchange = createMessageExchange(axisService.getConsumesCDKExtensions());
            exchangeWrapper = new ExchangeImpl(exchange);

            exchangeWrapper.setInMessageContent(source);
            exchangeWrapper.setOperation(operationName);

            // Set the service name, endpoint name and interface name. The
            // Router will find the JBI endpoint to contact with these
            // informations
            exchangeWrapper.setService(axisService.getConsumes().getServiceName());
            exchangeWrapper.setInterfaceName(axisService.getConsumes().getInterfaceName());
            exchangeWrapper.setEndpoint(this.findJBIEndpoint(axisService.getConsumes()));

            // add SOAP attachments to normalized message
            marshaller.copySOAPAttachmentsToNM(inContext.getAttachmentMap(), exchangeWrapper
                    .getInMessage());

            // retrieve security subject
            Subject subject = createSecuritySubject(inContext);
            if (subject != null) {
                exchangeWrapper.getInMessage().setSecuritySubject(subject);
            }

            // Get the options from the SOAP message and put them into the JBI
            // message
            marshaller.copySOAPPropertiesToME(inContext, exchangeWrapper);

            // Set the performance notification UID, if needed
            if (performanceNotification != null) {
                performanceNotification.setUIDInMessageExchange(exchange);
            }

        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }

        return exchangeWrapper.getMessageExchange();
    }

    /**
     * Create the security {@link Subject} from the {@link MessageContext}.
     * TODO : Check if there are security features in REST too.
     * 
     * @param messageContext
     * @return the security subject created from the message context if we have
     *         found {@link Principal} in the message context else return null;
     */
    @SuppressWarnings("unchecked")
    protected Subject createSecuritySubject(MessageContext messageContext) {
        this.logger.fine("Trying to create JAAS Security Subject");

        Subject subject = null;
        Vector wssec = (Vector) messageContext.getProperty(WSHandlerConstants.RECV_RESULTS);
        if (wssec != null) {
            for (Iterator iterator = wssec.iterator(); iterator.hasNext();) {
                WSHandlerResult wshandler = (WSHandlerResult) iterator.next();
                Vector wsSecEngineResults = wshandler.getResults();

                for (Object object : wsSecEngineResults) {
                    WSSecurityEngineResult wsresult = (WSSecurityEngineResult) object;

                    Principal principal = (Principal) wsresult
                            .get(WSSecurityEngineResult.TAG_PRINCIPAL);
                    if (principal != null) {
                        if (subject == null) {
                            subject = new Subject();
                        }

                        this.logger.fine("Got a principal from input message context : "
                                + principal.getName());
                        subject.getPrincipals().add(principal);
                    }
                }
            }
        }
        return subject;
    }

    /**
     * Get the operation from the message context. Assume that the operation is
     * always the root element of the SOAP body.
     * 
     * @param message
     * @return null if the operation has not been found
     */
    protected QName retrieveOperationFromSOAPBody(MessageContext message) {
        QName result = null;
        SOAPEnvelope env = message.getEnvelope();
        SOAPBody body = env.getBody();
        OMElement operation = body.getFirstElement();
        if (operation != null) {
            result = operation.getQName();
        }
        return result;
    }

    /**
     * Get the processor name. This name MUST be unique ie two processors can
     * not have the same name !
     * 
     * @return
     */
    protected abstract String getProcessorName();
}
