/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import static org.ow2.petals.binding.soap.Constants.SOAP.FAULT_CLIENT;
import static org.ow2.petals.binding.soap.Constants.SOAP.FAULT_SERVER;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;

import org.apache.axiom.soap.SOAPFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.MessageContext;
import org.ow2.petals.binding.soap.listener.SUPropertiesHelper;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.util.ExchangeImpl;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * 
 * Created on 14 févr. 08
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since
 * 
 */
public class SOAPProcessor extends RequestProcessor {

    /**
     * 
     * @param component
     * @param channel
     * @param log
     */
    public SOAPProcessor(AbstractComponent component, DeliveryChannel channel, Logger log) {
        super(component, channel, log);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.binding.soap.listener.incoming.processors.Processor#process()
     */
    @Override
    protected void process(final MessageContext inContext, final MessageContext outContext,
            final SOAPFactory factory, final PerformanceNotification performanceNotification)
            throws AxisFault {
        logger.log(Level.FINE, "Processing the incoming SOAP message");

        // Type of service has been checked before
        PetalsAxisService axisService = (PetalsAxisService) inContext.getAxisService();

        // Retrieve message operation and service
        QName jbiOperation = this.retrieveSOAPOperation(inContext);
        if (jbiOperation == null) {
            throw new AxisFault("Operation not found in message context", FAULT_CLIENT);
        }

        // get timeout
        long timeout = SUPropertiesHelper.retrieveTimeOut(axisService.getConsumesCDKExtensions()
                .getPetalsExtensions(), 0);

        // create the message exchange
        MessageExchange exchange = null;
        try {
            exchange = createMessageExchange(inContext, jbiOperation, axisService,
                    performanceNotification);
        } catch (AxisFault e) {
            String errorMsg = "Error while transforming SOAP request to JBI MessageExchange";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, FAULT_SERVER, e);
        }

        // send the message exchange through JBI
        // As the SOAP/HTTP request may wait for a response,
        // use the sendSynchrone MEP
        boolean sent = false;
        PerformanceNotification l_performanceNotification = null;
        try {
            // Performance notification creation
            l_performanceNotification = this.component.createPerformanceNotificationSendingToNMR(
                    inContext.getTo().getAddress(), new ExchangeImpl(exchange));

            sent = channel.sendSync(exchange, timeout);

        } catch (MessagingException e) {
            String errorMsg = "Error while sending message through JBI NMR.";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, FAULT_SERVER, e);
        } finally {
            // Performance notification sending
            this.component.sendPerformanceNotification(l_performanceNotification);
        }

        if (!sent) {
            throw new AxisFault("A timeout occurs calling the consumed service.", FAULT_SERVER);
        }

        // process the response received from JBI NMR. The out context is
        // updated
        this.processJBIResponse(exchange, outContext, factory);

    }

    /**
     * Return the operation according to the WSAAction SOAP attribute. If not
     * found in the WSA, get it from the SOAP body (root element)
     * 
     * @param inContext
     *            SOAPrequest context
     * @return operation as QName, null if no WSAAction is specified
     */
    protected QName retrieveSOAPOperation(MessageContext inContext) {
        QName operationName = null;
        String operation = inContext.getWSAAction();
        if (!StringHelper.isNullOrEmpty(operation)) {
            operationName = QName.valueOf(operation);
        } else {
            operationName = retrieveOperationFromSOAPBody(inContext);
        }
        return operationName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.binding.soap.listener.incoming.RequestProcessor#getProcessorName()
     */
    @Override
    protected String getProcessorName() {
        return "SOAP";
    }
}
