/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SoapListenerManager.java 154 27 sept. 06 alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import java.net.SocketException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.engine.Handler;
import org.apache.axis2.engine.Phase;
import org.apache.axis2.transport.http.server.HttpUtils;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.binding.soap.SoapProperties;
import org.ow2.petals.binding.soap.listener.incoming.jetty.SoapServer;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.ComponentConfiguration;
import org.ow2.petals.component.framework.su.ServiceUnitManager;

/**
 * This class is used to manage listeners for external WS addresses registered
 * during SU deployments
 *
 * @version $Rev$
 * @since Petals 1.1
 * @author alouis - EBM Websourcing
 * @author chamerling - EBM Websourcing
 */

public class SoapExternalListenerManager {

    protected Logger logger;

    protected Set<String> addresses;

    protected DeliveryChannel channel;

    protected AbstractComponent component;

    protected ServiceUnitManager bindingSUM;

    protected ConfigurationContext axisConfigurationContext;

    protected SoapServer httpServer;

    protected SoapServerConfig serverConfig;

    protected ComponentConfiguration componentConfiguration;

    protected SoapComponentContext soapContext;
    
    protected PetalsReceiver petalsReceiver;

    /**
     * Creates a new instance of {@link SoapExternalListenerManager}
     *
     * @param ccontext
     * @param dchannel
     * @param bindingSUM
     * @param axisConfigurationContext
     * @param soapContext
     * @param propertiesManager
     * @param log
     */
    public SoapExternalListenerManager(AbstractComponent component,
            DeliveryChannel dchannel, ServiceUnitManager bindingSUM,
            ConfigurationContext axisConfigurationContext,
            SoapComponentContext soapContext,
            ComponentConfiguration componentConfiguration,
            PetalsReceiver petalsReceiver, Logger log) {
        super();
        this.logger = log;
        this.addresses = new HashSet<String>();
        this.bindingSUM = bindingSUM;
        this.channel = dchannel;
        this.component = component;
        this.axisConfigurationContext = axisConfigurationContext;
        this.soapContext = soapContext;
        this.componentConfiguration = componentConfiguration;
        this.serverConfig = createServerConfig();
        this.petalsReceiver = petalsReceiver;
    }
    

    /**
     * Starts the listener manager : Start the embedded HTTP server and
     * initialize the dispatcher.
     *
     * @throws AxisFault
     */
    public void start() throws AxisFault {

        // start the axis http server
        startHttpServer();

        /*
         * The PetalsDispatcher object used by Axis is dynamicaly created by the
         * Axis http server. Some attributes have to be set to this object to
         * help it working well. The PetalsDispatcher object is retieved from
         * the axis2 http server configuration.
         */
        PetalsDispatcher petalsDispatcher = retrievePetalsDispatcherFromAxisConfiguration();
        petalsDispatcher.init(this.component, this.channel, this.bindingSUM, this.petalsReceiver,
                this.logger);
        
        this.logger.info("Component Information is available at " + serverConfig.buildBaseURL());
    }

    /**
     * Stops the listener manager : Stop the embedded HTTP server that handles
     * incoming requests.
     *
     * @throws AxisFault
     */
    public void stop() throws AxisFault {
        stopHttpServer();
    }

    /**
     * Init and start the HTTP server which will handle external SOAP request.
     * The axis2.xml defines a Petals dispatcher that will catch incoming
     * requests.
     *
     * @throws AxisFault
     */
    protected void startHttpServer() throws AxisFault {
        // start the Jetty server
        this.httpServer = new SoapServer(axisConfigurationContext,
                serverConfig, soapContext, logger);
        this.httpServer.start();
    }

    /**
     * Stop the HTTP server
     *
     */
    protected void stopHttpServer() throws AxisFault {
        this.httpServer.stop();
    }

    /**
     * Retrieve from the AxisConfiguration the PetalsDispatcher object.
     * (AxisConfiguration/GlobalInFlow/Dispatch/PetalsDispatcher)
     *
     * @return PetalsDispatcher object
     * @throws AxisFault
     *             PetalsDispatcher object not found
     *
     */
    protected PetalsDispatcher retrievePetalsDispatcherFromAxisConfiguration()
            throws AxisFault {

        PetalsDispatcher petalsDispatcher = null;

        List axisPhases = this.httpServer.getConfigurationContext()
                .getAxisConfiguration().getInFlowPhases();

        for (Iterator iter = axisPhases.iterator(); iter.hasNext()
                && petalsDispatcher == null;) {

            Phase phase = (Phase) iter.next();

            if (phase.getPhaseName().equalsIgnoreCase("Dispatch")) {

                for (Iterator iterator = phase.getHandlers().iterator(); iterator
                        .hasNext()
                        && petalsDispatcher == null;) {

                    Handler handler = (Handler) iterator.next();

                    if (handler.getName().toString().equalsIgnoreCase(
                            "PetalsDispatcher")) {
                        petalsDispatcher = (PetalsDispatcher) handler;
                    }
                }
            }
        }

        if (petalsDispatcher == null) {
            throw new AxisFault(
                    "The PetalsDispatcher object can not be retrieved from the AxisConfiguration.");
        }
        return petalsDispatcher;
    }

    /**
     * Create the SOAP server configuration
     *
     * @return
     */
    private SoapServerConfig createServerConfig() {
        SoapServerConfig soapConfig = new SoapServerConfig();
        SoapProperties cp = new SoapProperties(componentConfiguration);
        soapConfig.setPort(cp.getHttpPort());
        soapConfig.setProvidesList(cp.isProvidingServicesList());
        soapConfig.setProtocol(cp.getProtocol());
        soapConfig.setJettyThreadMaxPoolSize(cp.getHttpThreadMaxPoolSize());
        soapConfig.setJettyThreadMinPoolSize(cp.getHttpThreadMinPoolSize());
        soapConfig.setJettyAcceptors(cp.getHttpAcceptors());
        String localAddress = null;
        try {
            localAddress = HttpUtils.getIpAddress();
        } catch (SocketException e) {
        }
        if (localAddress == null) {
            localAddress = "127.0.0.1";
        }
        soapConfig.setHost(localAddress);
        soapConfig.setServicesContext(cp.getServicesContext());
        soapConfig.setServicesMapping(cp.getServicesMapping());
        return soapConfig;
    }

    /**
     * Get the URL where services can be accessed
     *
     * @return
     */
    public String getServicesURL() {
        return serverConfig.buildServicesURL();
    }

    public Set<String> getAddresses() {
        return addresses;
    }

}
